/*****************************************************************************
Copyright (c) 2008, Christopher D. Twigg

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.  
* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*******************************************************************************/


#include "BBLParser.h"

#include <string>
#include <iostream>
#include <typeinfo>

#include <boost/utility.hpp>
#include <boost/spirit/core.hpp>
#include <boost/spirit/attribute.hpp>
#include <boost/spirit/utility/chset.hpp>
#include <boost/spirit/utility/escape_char.hpp>
#include <boost/spirit/utility/confix.hpp>


using namespace boost::spirit;

class Token
{
public:
  Token(iterator_t itr) : location_(itr) {}
  virtual ~Token() {}
  iterator_t location() const { return location_; }
  virtual std::string toString() const = 0;

private:
  iterator_t location_;
};

class EOFToken : public Token
{
public:
  EOFToken( const iterator_t& itr ) : Token(itr) {}
  virtual ~EOFToken() {}
  std::string toString() const { return "<EOF>"; }
};

class PunctuationToken : public Token
{
public:
  enum PunctuationType
  {
    LEFT_BRACE,      // '{'
    RIGHT_BRACE,     // '}'
    BACKSLASH,       // '\'
    TWIDDLE,         // '~'
  };

  PunctuationToken( PunctuationType type, const iterator_t& itr )
    : Token(itr), punctuationType_(type) {}
  virtual ~PunctuationToken() {}
  PunctuationType type() const { return punctuationType_; }
  virtual std::string toString() const { return punctuationTypeToString( punctuationType_ ); }
  static std::string punctuationTypeToString( PunctuationType type )
  {
    switch( type )
    {
      case LEFT_BRACE: return "{";
      case RIGHT_BRACE: return "}";
      case BACKSLASH: return "\\";
      case TWIDDLE: return "~";
    }
  }

private:
  PunctuationType punctuationType_;
};

class StringToken : public Token
{
public: 
  StringToken( const std::string& val, const iterator_t& it )
    : Token(it), value_(val) {}
  std::string value() const { return value_; }
  virtual ~StringToken();
  virtual std::string toString() const { return value(); }

private:
  std::string value_;
};


class FileTokenStream : public boost::noncopyable
{
public:
  FileTokenStream( const std::string& filename )
    : first_(filename.c_str()), last_(first_.make_end()), 
      current_(first_, last_, filename)
  {
  }

  virtual ~FileTokenStream() {}

  boost::shared_ptr< Token > peek()
  {
    if( currentToken_.get() == 0 )
    {
      // First, we clear out all the junk with our skip_grammar
      {
        parse_info<iterator_t> info = parse( current_, end_, space_p );
        current_ = info.stop;
      }

      if( finished() )
        currentToken_.reset( new EOFToken(current_) );
      else if( *current_ == '~' )
      {
        currentToken_.reset( new PunctuationToken(
          PunctuationToken::TWIDDLE, current_ ) );
        ++current_;
      }
      else if( *current_ == '\\' )
      {
        currentToken_.reset( new PunctuationToken(
          PunctuationToken::BACKSLASH, current_ ) );
        ++current_;
      }
      else if( *current_ == '{' )
      {
        currentToken_.reset( new PunctuationToken(
          PunctuationToken::LEFT_BRACE, current_ ) );
        ++current_;
      }
      else if( *current_ == '}' )
      {
        currentToken_.reset( new PunctuationToken(
          PunctuationToken::RIGHT_BRACE, current_ ) );
        ++current_;
      }
      else
      {
        std::string str;
        while( !isspace(*current_) 
          && *current_ != '{' && *current_ != '}' 
          && *current_ != '\\' && *current_ != '~' )
          str.push_back( *current_++ );
        currentToken_.reset( new StringToken( str, current_ ) );
      }
    }
  }

  bool finished()
  {
    return current_ == end_;
  }

private:
  boost::shared_ptr< Token > currentToken_;

  file_iterator<> first_;
  file_iterator<> last_;

  iterator_t current_;
  iterator_t end_;
};


