/*****************************************************************************
Copyright (c) 2008, Christopher D. Twigg

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.  
* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*******************************************************************************/


#include "bibtexParser.h"
#include "exception.h"

#include <iostream>
#include <string>


#include <boost/utility.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/scoped_array.hpp>
#include <boost/tokenizer.hpp>
#include <boost/spirit/core.hpp>
#include <boost/spirit/attribute.hpp>
#include <boost/spirit/utility/chset.hpp>
#include <boost/spirit/utility/escape_char.hpp>
#include <boost/spirit/utility/confix.hpp>
#include <boost/spirit/iterator.hpp>

#include <cstdio>
#include <cerrno>

extern "C"
{
#include <btparse.h>
}

using namespace boost::spirit;

class BTParseTree : public boost::noncopyable
{
public:
  BTParseTree( AST* ast )
    : ast_(ast)
  {
  }

  ~BTParseTree()
  {
    bt_free_ast( ast_ );
  }

  AST* get()
  {
    return ast_;
  }

private:
  AST* ast_;
};

class BTParser : public boost::noncopyable
{
public:
  BTParser( const std::string& filename )
    : filename_(filename),
      filePtr_( fopen( filename.c_str(), "r" ) )
  {
    if( filePtr_ == NULL )
      throw ParserException( strerror(errno) );

    filenameForBT_.reset( new char[filename_.size() + 1] );
    strcpy( filenameForBT_.get(), filename_.c_str() );

    bt_initialize();

    // Define month macros
    // jan, feb, mar, apr, may, jun, jul, aug, sep, oct, nov, dec
    bt_add_macro_text( "jan", "1", filenameForBT_.get(), 0 );
    bt_add_macro_text( "feb", "2", filenameForBT_.get(), 0 );
    bt_add_macro_text( "mar", "3", filenameForBT_.get(), 0 );
    bt_add_macro_text( "apr", "4", filenameForBT_.get(), 0 );
    bt_add_macro_text( "may", "5", filenameForBT_.get(), 0 );
    bt_add_macro_text( "jun", "6", filenameForBT_.get(), 0 );
    bt_add_macro_text( "jul", "7", filenameForBT_.get(), 0 );
    bt_add_macro_text( "aug", "8", filenameForBT_.get(), 0 );
    bt_add_macro_text( "sep", "9", filenameForBT_.get(), 0 );
    bt_add_macro_text( "oct", "10", filenameForBT_.get(), 0 );
    bt_add_macro_text( "nov", "11", filenameForBT_.get(), 0 );
    bt_add_macro_text( "dec", "12", filenameForBT_.get(), 0 );

    parseNextEntry();

  }

  bool hasNext()
  {
    return (next_->get() != 0);
  }

  boost::shared_ptr<BTParseTree> next()
  {
    boost::shared_ptr<BTParseTree> res = next_;
    parseNextEntry();
    return res;
  }

  ~BTParser()
  {
    bt_cleanup();
    fclose( filePtr_ );
  }

private:
  void parseNextEntry()
  {
    boolean ok;
    ushort options = 0;
    next_.reset( new BTParseTree(
      bt_parse_entry(filePtr_, filenameForBT_.get(), options, &ok ) ) );
    if( next_->get() != 0 && !ok )
      throw ParserException( "Unknown error while parsing BiBTeX file." );
  }
  std::string filename_;
  boost::scoped_array<char> filenameForBT_;
  FILE* filePtr_;

  boost::shared_ptr<BTParseTree> next_;
};

std::string dump_component( bt_name* name, int part )
{
  std::string result;
  char **tokens = name->parts[part];
  int num_tokens = name->part_len[part];
  for( int i = 0; i < num_tokens; ++i )
  {
    if( !result.empty() )
      result.append( " " );
    result.append( tokens[i] );
  }

  return result;
}

Author::Author( bt_name* name )
{
  first_ = dump_component( name, BTN_FIRST );
  von_ = dump_component( name, BTN_VON );
  last_ = dump_component( name, BTN_LAST );
  jr_ = dump_component( name, BTN_JR );
}

std::vector<Author> authors( const char* s )
{
  std::vector<Author> result;
  std::string ss(s);
  bt_stringlist* namelist = bt_split_list (&ss[0], "and", NULL, 0, "name");
  for( int i=0; i < namelist->num_items; i++)
    result.push_back( 
      Author( 
        bt_split_name( namelist->items[i], NULL, 0, i ) ) );
  return result;
}

BTEntry::BTEntry( AST* entry )
{
  AST* field = NULL;
  char* fname;
  key_ = bt_entry_key(entry);
  type_ = bt_entry_type(entry);

  while ((field = bt_next_field (entry, field, &fname)))
  {
    fields_.push_back( 
      std::make_pair( 
        std::string(fname), 
        std::string(bt_get_text(field)) ) );

    std::string fieldName( fname );
    if( fieldName == "author" )
    {
      char* value = bt_get_text(field);
      authors_ = ::authors( value );
    }
    else if( fieldName == "journal" )
    {
      journal_ = bt_get_text(field);
    }
    else if( fieldName == "booktitle" )
    {
      bookTitle_ = bt_get_text(field);
    }
    else if( fieldName == "title" )
    {
      title_ = bt_get_text(field);
    }
    else if( fieldName == "year" )
    {
      char* value = bt_get_text(field);
      parse_info<char const*> info = parse( value, 
        uint_p[ assign(year_) ], space_p );
      if( !info.hit )
        throw ParserException( "Unable to read year in entry '" + key_ + "'." );
    }
    else if( fieldName == "month" )
    {
      char* value = bt_get_text(field);
      parse( value, uint_p[ assign(month_) ], space_p );
      // ignore misses; month will be 0.
    }
    else if( fieldName == "abstract" )
    {
      abstract_ = bt_get_text(field);
    }
    else if( fieldName == "image" )
    {
      image_ = bt_get_text(field);
    }
    else if( fieldName == "sidebarimage" )
    {
      sidebarImage_ = bt_get_text(field);
    }
    else if( fieldName == "iconimage" )
    {
      iconImage_ = bt_get_text(field);
    }
    else if( fieldName == "sidebartext" )
    {
      sidebarText_ = bt_get_text(field);
    }
    else if( fieldName == "links" )
    {
      typedef boost::tokenizer<boost::char_separator<char> > tokenizer;
      boost::char_separator<char> sep("; ");
      std::string text( bt_get_text(field) );
      tokenizer tokens(text, sep);
      for (tokenizer::iterator tok_iter = tokens.begin();
         tok_iter != tokens.end(); ++tok_iter)
      {
        links_.push_back( *tok_iter );
      }
    }
  }
}


std::string BTEntry::bibEntry() const
{
  std::ostringstream oss;
  oss << "@" << type_ << "{ " << key_ << "," << std::endl;
  for( std::vector<StringPair>::const_iterator itr = fields_.begin();
    itr != fields_.end();
    ++itr )
  {
    std::string name = itr->first;
    std::string value = itr->second;
    if( itr->first == "abstract" || 
        itr->first == "image" || 
        itr->first == "iconimage" || 
        itr->first == "links" || 
        itr->first == "sidebartext" || 
        itr->first == "sidebarimage" )
      continue;

    if( itr->first == "month" && month_ >= 1 && month_ <= 12 )
    {
      switch( month_ ) 
      {
        case 1:  value = "jan"; break;
        case 2:  value = "feb"; break;
        case 3:  value = "mar"; break;
        case 4:  value = "apr"; break;
        case 5:  value = "may"; break;
        case 6:  value = "jun"; break;
        case 7:  value = "jul"; break;
        case 8:  value = "aug"; break;
        case 9:  value = "sep"; break;
        case 10:  value = "oct"; break;
        case 11:  value = "nov"; break;
        case 12:  value = "dec"; break;
      }
      oss << "  " << name << " = " << value << "," << std::endl;
    }
    else
      oss << "  " << name << " = \"" << value << "\"," << std::endl;
  }
  oss << "}" << std::flush;

  return oss.str();
}

std::vector<BTEntry> parseBiBTeX( const std::string& filename )
{
  BTParser parser( filename );
  std::vector<BTEntry> entries;

  while( parser.hasNext() )
  {
    boost::shared_ptr<BTParseTree> ptr = parser.next();
    if( bt_entry_metatype(ptr->get()) != BTE_REGULAR )
    {
      std::cerr << "Skipping " << bt_entry_type(ptr->get()) << " entry."
        << std::endl;
      continue;
    }

    entries.push_back( BTEntry( ptr->get() ) );
  }

  return entries;
}



