/*****************************************************************************
Copyright (c) 2008, Christopher D. Twigg

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.  
* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*******************************************************************************/


#include "httpWrapper.h"
#include "exception.h"

#include <curl/curl.h>
#include <cstring>
#include <cstdio>
#include <cctype>

#include <boost/spirit/core.hpp>
#include <boost/spirit/attribute.hpp>
#include <boost/spirit/utility/chset.hpp>
#include <boost/spirit/utility/escape_char.hpp>
#include <boost/spirit/utility/confix.hpp>
#include <boost/spirit/iterator.hpp>

using namespace std;
using namespace boost::spirit;

size_t write_header( void *ptr, size_t size, size_t nmemb, void *stream );
size_t write_data( void *ptr, size_t size, size_t nmemb, void *stream );

class CURLWrapper
{
public:
  CURLWrapper( const std::string& url, bool fetchData = false )
    : url_(url)
  {
    char errbuf[ CURL_ERROR_SIZE+1 ];

    curl_ = curl_easy_init();
    curl_easy_setopt(curl_, CURLOPT_TIMEOUT, 10);
    curl_easy_setopt(curl_, CURLOPT_NOPROGRESS, 1);
    curl_easy_setopt(curl_, CURLOPT_NOPROGRESS, 1);
//    curl_easy_setopt(curl_, CURLOPT_HEADER, 1);
//    curl_easy_setopt(curl_, CURLOPT_WRITEFUNCTION, write_data );
    curl_easy_setopt(curl_, CURLOPT_ERRORBUFFER, errbuf );
    curl_easy_setopt(curl_, CURLOPT_HEADERFUNCTION, write_header );
    curl_easy_setopt(curl_, CURLOPT_URL, url_.c_str() );
    curl_easy_setopt(curl_, CURLOPT_WRITEHEADER, (void *)this);

    if( fetchData )
    {
      curl_easy_setopt(curl_, CURLOPT_WRITEFUNCTION, write_data );
      curl_easy_setopt(curl_, CURLOPT_WRITEDATA, (void *)this);
    }
    else
    {
      curl_easy_setopt(curl_, CURLOPT_NOBODY, 1);
    }

    CURLcode res = curl_easy_perform(curl_);
    if( res != 0 )
      throw IOException( errbuf );
  }

  ~CURLWrapper()
  {
    curl_easy_cleanup(curl_);
  }

  const HttpResponse response() const { return response_; }
  std::string contentType() const { return response_.contentType(); }
  unsigned long size() const { return response_.size(); }

private:
  CURL* curl_;
  std::string url_;
  HttpResponse response_;

  friend size_t write_data( void *ptr, size_t size, size_t nmemb, void *stream );
  friend size_t write_header( void *ptr, size_t size, size_t nmemb, void *stream );
};

std::string toLowerCase( const char* cur, const char* end )
{
  std::string res;
  res.reserve( end - cur );
  while( cur != end )
    res.push_back( tolower(*cur++) );
  return res;
}

size_t write_data( void *ptr, size_t size, size_t nmemb, void *stream )
{
  register int realsize = size * nmemb;
  const char* array = (const char*) ptr;
  CURLWrapper* wrapper = (CURLWrapper*) stream;

  wrapper->response_.data_.reserve( wrapper->response_.data_.size() + realsize );
  std::copy( array, array + realsize, std::back_inserter(wrapper->response_.data_) );

  return realsize;
}

size_t write_header( void *ptr, size_t size, size_t nmemb, void *stream )
{
  register int realsize = size * nmemb;
  const char* array = (const char*) ptr;
  CURLWrapper* wrapper = (CURLWrapper*) stream;

  parse_info<char const*> info = parse( array, array + realsize, 
    lexeme_d[ as_lower_d[str_p( "http/" ) >> +(digit_p | ch_p('.'))] ], space_p );
  if( info.hit )
  {
    unsigned int errCode;
    info = parse( info.stop, array + realsize, 
      uint_p[ assign_a( errCode ) ], space_p );
    if( info.hit )
      wrapper->response_.errorCode_ = errCode;
  }

  info = parse( array, array + realsize, 
    alnum_p >> *(alnum_p | ch_p('-')) >> ch_p(':') );
  if( info.hit )
  {
    std::string fieldName( toLowerCase(array, info.stop - 1) );
    std::string fieldValue( toLowerCase(info.stop, array + realsize) );

    const char* valueStart = info.stop;

    if( fieldName == "content-length" )
    {
      unsigned long length;
      parse_info<char const*> info = parse( valueStart, array + realsize, 
        uint_p[ assign(length) ], space_p );
      if( info.hit )
        wrapper->response_.size_ = length;
    }
    else if( fieldName == "content-type" )
    {
      // should do something neat here.
      while( isspace(*valueStart) && valueStart != array + realsize ) ++valueStart;
      parse_info<char const*> info = 
        parse( valueStart, array + realsize, +graph_p );

      if( info.hit )
      {
        std::string contentType( valueStart, info.stop );
        wrapper->response_.contentType_ = contentType;
      }
    }
  }

  return realsize;
}

const HttpResponse getHeader( const std::string& url )
{
  CURLWrapper wrapper( url );
  return wrapper.response();
}

/*
ILImageWrapper::ILImageWrapper()
{
  ILuint images[1];
  ilGenImages( 1, images );
  image_ = images[0];
}

ILImageWrapper::~ILImageWrapper()
{
  ILuint images[1] = { image_ };
  ilDeleteImages( 1, images );
}

ILuint ILImageWrapper::imageId() const
{
  return image_;
}

ILBindImage::ILBindImage( ILuint imageId )
{
  ilBindImage( imageId );
}

ILBindImage::~ILBindImage()
{
  ilBindImage( 0 );
}
*/

ImageResponse::ImageResponse( HttpResponse resp, const Magick::Image& image )
  : HttpResponse( resp ), image_(image)
{
}

unsigned int ImageResponse::width() const
{
	return image_.size().width();
}

unsigned int ImageResponse::height() const
{
	return image_.size().height();
}

const Magick::Image& ImageResponse::image() const
{
	return image_;
}

const ImageResponse getImage( const std::string& url )
{
	// we'll just do a header fetch first to make sure the URL
	//  is okay.
  CURLWrapper wrapper( url, true );
  HttpResponse response( wrapper.response() );
	if( response.errorCode() >= 400 )
		return ImageResponse( response, Magick::Image() );

  // Now, need to feed the image to DevIL
  std::string contentType = response.contentType();

  /*
  std::string magick;
  if( contentType.find( "jpeg" ) != std::string::npos )
    magick = "JPEG";
  else if( contentType.find( "gif" ) != std::string::npos )
    magick = "GIF";
  else if( contentType.find( "png" ) != std::string::npos )
    magick = "PNG";
  else
    throw IOException( "Unknown content type for image: " + contentType );
  */

	try
	{
    std::vector<char> lump( response.data() );
    Magick::Blob blob( &lump[0], response.size() );
		Magick::Image image( blob );
		return ImageResponse( response,
			image );
	}
	catch( Magick::Exception& e )
	{
		throw IOException( "Error reading image from URL '" + url + "': " + e.what() );
	}

/*

  {
    static bool il_initialized = false;
    if( !il_initialized )
    {
      ilInit();
      iluInit();
      il_initialized = true;
    }
  }


  // Now, need to feed the image to DevIL
  std::string contentType = response.contentType();

  ILenum type;
  if( contentType.find( "jpeg" ) != std::string::npos )
    type = IL_JPG;
  else if( contentType.find( "gif" ) != std::string::npos )
    type = IL_GIF;
  else if( contentType.find( "png" ) != std::string::npos )
    type = IL_PNG;
  else
    type = IL_TYPE_UNKNOWN;

  boost::shared_ptr<ILImageWrapper> ilImageWrapper( new ILImageWrapper );
  ILBindImage bindImage( ilImageWrapper->imageId() );
  std::vector<char> lump( response.data() );
  ILboolean ilResponse = ilLoadL( type, (ILvoid*) &lump[0], lump.size() );
  if( !ilResponse )
  {
    throw IOException( std::string("Unable to read image '") + url + std::string("': ")
      + std::string(iluErrorString(ilGetError())) );
  }

  return ImageResponse( response,
    ilGetInteger( IL_IMAGE_WIDTH ), 
    ilGetInteger( IL_IMAGE_HEIGHT ),
		ilImageWrapper );
*/
}


