/*****************************************************************************
Copyright (c) 2008, Christopher D. Twigg

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.  
* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*******************************************************************************/


#include "latexParser.h"
#include "exception.h"
#include "htmlWrappers.h"

#include <iostream>
#include <stack>
#include <sstream>
#include <vector>
#include <boost/shared_ptr.hpp>
#include <boost/scoped_ptr.hpp>
#include <boost/utility.hpp>
#include <boost/spirit/iterator.hpp>

using namespace std;
using namespace boost::spirit;

std::string HtmlAccented( char in, char accentType )
{
  if( accentType == '\'' )
  {
    if( in == 'c' )
    {
      // for the Popovic brothers
      return std::string("&#263;");
    }
  }
  else if( accentType == 'v' )
  {
    if( in == 'c' )
    {
      // For Jernej
      return std::string("&#269;");
    }
  }
  else if( accentType == 'c' )
  {
    if ( in == 'c' )
    {
      return std::string("&ccedil;");
    }
  }

  std::string res;
  res.push_back(in);
  return res;
}

class BBLParser : boost::noncopyable
{
  class ItemAppender
  {
  public:
    ItemAppender( BBLParser& parser, const std::string& name = std::string() )
      : bblParser_(parser), name_(name) {}

    ~ItemAppender()
    {
      if( !name_.empty() )
         bblParser_.entries_[name_] = text_.str();

      if( !bblParser_.fullText_.empty() )
        bblParser_.fullText_.push_back(' ');
      bblParser_.fullText_.append( text_.str() );
    }

    template <typename T>
    ItemAppender& operator<<( const T& rhs )
    {
      text_ << rhs;
      return *this;
    }

    std::string str() const { return text_.str(); }

    operator ostream&() { return text_; }

  private:
    BBLParser& bblParser_;
    std::string name_;
    std::ostringstream text_;
  };

public:
  template <typename Iterator>
  BBLParser( Iterator& current, const Iterator& end )
  {
    typedef std::vector<boost::shared_ptr<HtmlThing> > Environment;
    std::stack< Environment > environments;

    boost::scoped_ptr<ItemAppender> appender(
      new ItemAppender( *this ) );
    while( current != end )
    {
      if( isspace( *current ) )
      {
        if( !isspace( *(*appender).str().rbegin() ) )
          (*appender) << " ";
        ++current;
      }
      else if( *current == '{' )
      {
        environments.push( Environment() );
        ++current;
      }
      else if( *current == '}' )
      {
        if( environments.empty() )
          throw ParserException( "Unmatched '}'", current );
        environments.pop();
        ++current;
      }
      else if( *current == '<' )
      {
        (*appender) << "&lt;";
        ++current;
      }
      else if( *current == '>' )
      {
        (*appender) << "&gt;";
        ++current;
      }
      else if( *current == '\\' )
      {
        std::string escaped = parseEscaped( current, end );
        if( escaped == "bibitem" )
        {
          appender.reset( new ItemAppender( *this, parseWrapped( current, end ) ) );
        }
        else if( escaped == "newblock" )
        {
          (*appender) << " ";
        }
        else if( escaped == "em" )
        {
          environments.top().push_back( 
            boost::shared_ptr<HtmlThing>( 
              new HtmlThing((*appender), "i") ) );
        }
        else if( escaped == "begin" )
        {
          std::string beginWhat = parseWrapped( current, end );
          if( beginWhat != "thebibliography" )
            throw ParserException( "Invalid 'begin'", current );
        }
        else if( escaped == "end" )
        {
          std::string endWhat = parseWrapped( current, end );
          if( endWhat != "thebibliography" )
            throw ParserException( "Invalid 'end'", current );
        }
        else if( escaped == "&" )
        {
          (*appender) << "&amp;";
        }
        else if( escaped == "'" )
        {
          (*appender) << parseAccented( current, end, '\'' );
        }
        else if( escaped == "c" )
        {
          (*appender) << parseAccented( current, end, 'c' );
        }
        else if( escaped == "v" )
        {
          (*appender) << parseAccented( current, end, 'v' );
        }
        else
          throw ParserException( "Unknown escaped word '" + escaped + "'.", current );
      }
      else if( *current == '~' )
      {
        (*appender) << " ";
        current++;
      }
      else if ( *current == '-' )
      {
        unsigned int dash = 0;
        while( *current == '-' )
        {
          ++current;
          ++dash;
        }

        if( dash == 1 )
          (*appender) << "-";        // hyphen
        else if( dash == 2 )
          (*appender) << "&ndash;";  // en dash
        //  (*appender) << "&#8211;";  // en dash
        else if( dash == 3 )
          (*appender) << "&#8212;";  // em dash
        else 
          throw ParserException( "Too many dashes in a row.", current );
      }
      else
      {
        (*appender) << *current++;
      }
    }
  }

  std::string citation( const std::string& key ) const
  {
    CitationMap::const_iterator itr = entries_.find( key );
    if( itr == entries_.end() )
      throw ParserException( "Missing citation for '" + key + "'." );
    return itr->second;
  }

  CitationMap map() const
  {
    return entries_;
  }

  std::string fullText() const
  {
    return fullText_;
  }

private:
  template <typename Iterator>
  std::string parseWrapped( Iterator& current, const Iterator& end ) const
  {
    std::string res;
    parsePunct( '{', current, end );
    while( *current != '}' && current != end )
    {
      if( !isspace(*current) )
         res.push_back(*current);
      ++current;
    }
    parsePunct( '}', current, end );
    return res;
  }

  template <typename Iterator>
  std::string parseEscaped( Iterator& current, const Iterator& end ) const
  {
    ++current;                // eat the '\' character
    if( current == end )
      throw ParserException( "Unterminated escaped char.", current );

    if( ispunct(*current) )
    {
      std::string res;
      res.push_back( *current++ );
      return res;
    }
    else if( isalpha( *current ) )
    {
      std::string res;
      while( current != end && isalpha(*current) )
        res.push_back(*current++);
      return res;
    }
    else
    {
      throw ParserException( "Unknown escaped character", current );
    }
  }

  template <typename Iterator>
  void parsePunct( char punct, Iterator& current, const Iterator& end ) const
  {
    while( isspace( *current ) && current != end ) ++current;
    if( current == end || *current != punct )
      throw ParserException( "Invalid punctuation", current );
    ++current;
    return;
  }

  template <typename Iterator>
  std::string parseAccented( Iterator& current, const Iterator& end, char accentType ) const 
  {
    std::string res;
    parsePunct( '{', current, end );
    while( *current != '}' && current != end )
    {
      res.append( HtmlAccented( *current, accentType ) );
      current++;
    }
    parsePunct( '}', current, end );
    return res;
  }

  CitationMap entries_;
  std::string fullText_;
};

std::string formatLatexAsHTML( const std::string& in )
{
  std::string::const_iterator begin = in.begin();
  std::string::const_iterator end = in.end();
  BBLParser bbl(begin, end);
  return bbl.fullText();
}

CitationMap parseBBLFile( const std::string& filename )
{
  file_iterator<> first( filename.c_str() );
  if( !first )
  {
    std::string result;
    result.append( "Unable to open file: '" );
    result.append( filename );
    result.append( "' for reading." );
    throw IOException( result );
  }

  file_iterator<> last( first.make_end() );
  iterator_t begin( first, last, filename );
  iterator_t end;
  BBLParser bbl( begin, end );

  return bbl.map();
}


